#pragma once
#include "context.h"
#include "buffer.h"

[[nodiscard]] std::vector<wgpu::BindGroupLayout> createBindGroupLayouts(const wgpu::Device& device, const Shader& shader, wgpu::ShaderStage shaderStage = wgpu::ShaderStage::Compute);

[[nodiscard]] std::vector<wgpu::BindGroupLayout> createBindGroupLayouts(const wgpu::Device& device, const std::vector<std::vector<wgpu::BindGroupLayoutEntry>>& entries);

template<typename T>
inline void fillEntry(wgpu::BindGroupEntry& entry, const T& input) {
	static_assert(false, "type not supported");
}

template<>
inline void fillEntry<Buffer>(wgpu::BindGroupEntry& entry, const Buffer& buffer) {
	entry.buffer = buffer;
	entry.size = buffer.desc.size;
}

template<>
inline void fillEntry<wgpu::Sampler>(wgpu::BindGroupEntry& entry, const wgpu::Sampler& sampler) {
	entry.sampler = sampler;
}

template<>
inline void fillEntry<wgpu::TextureView>(wgpu::BindGroupEntry& entry, const wgpu::TextureView& view) {
	entry.textureView = view;
}

template<>
inline void fillEntry<BufferSubrange>(wgpu::BindGroupEntry& entry, const BufferSubrange& subrange) {
	entry.buffer = subrange.buffer();
	entry.size = subrange.size();
	entry.offset = subrange.offset();
}

template<>
inline void fillEntry<nullptr_t>(wgpu::BindGroupEntry&, const nullptr_t&) {}


template<typename ...Entries>
inline wgpu::BindGroup createBindGroup(wgpu::Device device, wgpu::BindGroupLayout layout, 
	const Entries&... bindEntries) 
{
	constexpr std::size_t args_count = sizeof...(Entries);
	std::vector<wgpu::BindGroupEntry> entries;
	entries.reserve(args_count);
	unsigned i = 0;
	const auto addEntry = [&](const auto& input) {
		if constexpr (!std::same_as<decltype(input), const nullptr_t&>) { // nullptr means skip binding i
			wgpu::BindGroupEntry entry{ .binding = i };
			fillEntry(entry, input);
			entries.push_back(std::move(entry));
		}
		i++;
	};
	(addEntry(bindEntries), ...);

	const wgpu::BindGroupDescriptor desc{
		.layout = layout,
		.entryCount = static_cast<uint32_t>(entries.size()),
		.entries = entries.data(),
	};
	return device.CreateBindGroup(&desc);
}