#pragma once

#include <concepts>
#include <cstdint>
#include <cstddef>
#include <type_traits>
#include <atomic>

template<typename ContType>
concept Container = requires(ContType a) {
	{ a.size() } -> std::convertible_to<size_t>;
	{ a.data() } -> std::convertible_to<typename ContType::value_type*>;
};

template<Container ContType>
inline size_t byteSize(const ContType& c) {
	return c.size() * sizeof(ContType::value_type);
}

inline const std::byte* asBytes(const Container auto& c) {
	return reinterpret_cast<const std::byte*>(c.data());
}

inline const std::byte* asBytes(const auto* ptr) {
	return reinterpret_cast<const std::byte*>(ptr);
}

inline std::byte* asWritableBytes(Container auto& c) {
	return reinterpret_cast<std::byte*>(c.data());
}

inline std::byte* asWritableBytes(auto* ptr) {
	return reinterpret_cast<std::byte*>(ptr);
}

inline constexpr auto ceilDivide(std::integral auto val, std::integral auto div) {
	return (val + div - 1) / div;
}

// Round up val to a nearest multiple of a.
template <typename T>
constexpr inline T roundUp(T val, auto a)
{
	return T(((val + a - 1) / a) * a);
}
