#pragma once
#include "dawn/dawn_proc.h"
#include "dawn_native/D3D12Backend.h"
#include "dawn_native/VulkanBackend.h"
#include "dawn/webgpu_cpp.h"
#undef min
#undef max

#include <filesystem>
#include <string>

const inline uint32_t bufferSubrangeAllignment = 256;

struct GLFWwindow;

void onMappingDone(WGPUBufferMapAsyncStatus status, void* flagVoidPtr);
void onQueueDone(WGPUQueueWorkDoneStatus status, void* flagVoidPtr);

class Shader {
public:
	wgpu::ShaderModule compiled;
	std::string source;

	Shader() = default;
	Shader(const wgpu::Device& device, std::filesystem::path filePath);
	void discardSource() {
		source.clear();
	}
	operator wgpu::ShaderModule& () {
		return compiled;
	}
};
class Buffer;

class Context {
public:
	wgpu::Device device;
	wgpu::Queue queue;
	mutable bool emptyEncoder = true;
	mutable wgpu::CommandEncoder encoder;

	wgpu::Texture renderImage;
	wgpu::Extent3D imageSize;

private:
	bool glfwInitialised = false;
	GLFWwindow*  window = nullptr;
	
	DawnProcTable procTable;

	wgpu::Surface surface;
	wgpu::SwapChain swapchain;
	
	std::unique_ptr<Buffer> windowSizeUniform;
	wgpu::Sampler renderSampler;
	wgpu::BindGroupLayout renderBindGroupLayout;
	Shader vertexShader;
	Shader fragmentShader;
	wgpu::RenderPipeline renderPipeline;

	void initGLFW();
public:
	Context(bool createWindow = true);
	~Context();

	Context(const Context& other) = delete;
	Context& operator=(const Context& other) = delete;
	Context(Context&& other) = delete;
	Context& operator=(Context&& other) = delete;
	
	bool shouldClose();

	void pollEvents();

	void drawRenderTexture();

	void flushCommands() const;

	void flushCommandsAndWait() const {
		flushCommands();
		std::atomic_flag done;
		queue.OnSubmittedWorkDone(0, onQueueDone, &done);
		while (done.test()) {
			device.Tick();
		}
	}
};